/*------------------------------------------
   WH_WEDIT.C -- WINHELP File Editor
   (c) 1994, 1995 Bent Lynggaard
  ------------------------------------------*/

#include <windows.h>
#include <commdlg.h>
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <io.h>
#include <mem.h>
#include <string.h>
#include <fcntl.h>
#include "help_bas.h"
#include "wh_wedit.h"

#define PROGNAME "WH_wEdit"
#define MAX_WINDOWS 10
  // Windows v. 3.1 supports only 6

#define elements(array) (sizeof(array)/sizeof(array[0]))

typedef struct {
	WORD			offset;
	int				length;
	BOOL			secondary;
	int				changed;
	WORD			flag;
	LONG			pos;
	LONG			size;
	COLORREF		rgbNSC;
	COLORREF		rgbSC;
	WINDOWSTRUCT	w;
} EXT_WS; // typedef struct

static LOCALHANDLE ewpWindows[MAX_WINDOWS];

#define PSDEF	1
#define PSOK	2
#define HDEF	4
#define HOK		8
#define TDEF	0x10
#define TOK		0x20
#define EDITID	1
#define SAVEWINDOW	1
#define SAVEMACRO	2

long CALLBACK _export WndProc(HWND, UINT, UINT, LONG);
long CALLBACK _export EditProc(HWND, UINT, UINT, LONG);

static int		iCurrentWindow = -1;
static int		cXscreen;
static int		cYscreen;
static int		cYcaption;
static int		cXdlgframe;
static int		cYdlgframe;
static int		cYmenu;
static int		cXoffs;
static int		cYoffs;
static int		cXadj;
static int		cYadj;
static int		cScanmark;
static int		cXorig;
static int		cYorig;
static int		cWidth;
static int		cHeight;
static char		szAppName[] = PROGNAME;
static char		szEditFrame[] = PROGNAME "_EMF";
static char		szFileName[_MAX_PATH];
static char		szTitleName[_MAX_FNAME + _MAX_EXT];
static HWND		hwndMain, hwndEditFrame, hwndEdit;
static HWND		hwndHeaderSample, hwndTextSample;
static HANDLE	hInst;
static int		hFile;
static int		nCount;		// Number of windows in file
static int		iNeedSave, iSave;
static BOOL		bFileLoaded;
static BOOL		bSampleMode;
static BOOL		bEditEnabled;
static BOOL		bStrangeMacro;
static UINT		idColorHelp;
static DWORD	dwSysStart, dwSysMacros, dwSysWindows, dwSysEnd, dwSysLength;
EXT_WS			*ewp;		// Pointer to current extended w-struct
WINDOWSTRUCT	*wsp;		// Pointer to current window structure
WINHELPHEADER	root;
static char		buffer[24];	// General buffer
static FARPROC	lpfnAboutDlgProc;
static FARPROC	lpfnWH_wSampleProc;
static UINT textWindows[] = {IDT_FILENAME, IDT_COMPILED,
		IDT_NAME, IDT_KIND, IDT_TITLE, IDT_POS_SIZE,
		IDT_HEADER_COLOR, IDT_TEXT_COLOR};
static UINT cbWindows[] = {IDC_POS_SIZE, IDC_HEADER_COLOR, IDC_TEXT_COLOR};
static UINT toHide[] = {IDT_HEADER_SAMPLE, IDT_TEXT_SAMPLE,
		IDC_POS_SIZE_CHANGE, IDC_HEADER_CHANGE, IDC_TEXT_CHANGE};
static UINT toEnable[] = {IDL_WINDOW, IDC_POS_SIZE,
		IDC_HEADER_COLOR, IDC_TEXT_COLOR};
#pragma option -a-
static char	szFilterHelp[] = "Help Files (*.HLP)",  szFilterHLP[] = "*.hlp",
			szFilterEnd[] = "";
#pragma option -a.
static OPENFILENAME ofn = {
	/* lStructSize */		sizeof (OPENFILENAME),
	/* hwndOwner */			0,		// Set during initialization
	/* hInstance */			NULL,
	/* lpstrFilter */		(LPCSTR)&szFilterHelp,
	/* lpstrCustomFilter */	NULL,
	/* nMaxCustFilter */	0,
	/* nFilterIndex */		0,
	/* lpstrFile */			szFileName,
	/* nMaxFile */			_MAX_PATH,
	/* lpstrFileTitle */	szTitleName,
	/* nMaxFileTitle */		_MAX_FNAME + _MAX_EXT,
	/* lpstrInitialDir */	NULL,
	/* lpstrTitle */		NULL,
	/* Flags */				0,		// Set in Open and Close functions
	/* nFileOffset */		0,
	/* nFileExtension */	0,
	/* lpstrDefExt */		"hlp",
	/* lCustData */			0L,
	/* lpfnHook */			NULL,
	/* lpTemplateName */	NULL
}; // static OPENFILENAME ofn =
static COLORREF rgbUser[16] = {
	RGB(255,255,192),
	RGB(255,255,224),
	RGB(192,255,255),
	RGB(224,255,255),
	RGB(192,255,192),
	RGB(224,255,224),
	RGB(255,192,192),
	RGB(255,224,224),
	RGB(0,0,0),
	RGB(64,64,64),
	RGB(128,128,128),
	RGB(160,160,160),
	RGB(192,192,192),
	RGB(224,224,224),
	RGB(255,255,255),
	0L /* set to current color */
}; // static COLORREF rgbUser[16] =
static CHOOSECOLOR cc = {
	sizeof(CHOOSECOLOR),
	NULL,		// hwnd during initialization
	NULL,
	0L,
	rgbUser,	// user color when called
	CC_SHOWHELP | CC_RGBINIT,
	0L,
	NULL,
	NULL
}; // static CHOOSECOLOR cc =
static TEXTMETRIC tm;
static LOGFONT logfont;

void okMessage(int nMessage)
{
	char *pszBuffer;
	if ((pszBuffer = (char NEAR *)LocalAlloc(LMEM_FIXED,
			200+_MAX_FNAME+_MAX_EXT))!=NULL) {
		LoadString(hInst, nMessage, (LPSTR)pszBuffer, 100);
		pszBuffer[99] = 0;
		wsprintf((LPSTR)(pszBuffer+100), (LPSTR)pszBuffer, (LPSTR)szTitleName);
		MessageBox(hwndMain, pszBuffer+100, szAppName, MB_OK | MB_ICONEXCLAMATION);
		LocalFree((LOCALHANDLE)pszBuffer);
	}
} // void okMessage(int nMessage)

char *allocateAndLoad(int cbSize, UINT idStringNo)
{
#define ASTASTAST ('*' + '*' * 0x100 + '*' * 0x10000)
	char *pszBuffer, *pszLine;
	int n;
	if ((pszLine = pszBuffer = (char NEAR *)LocalAlloc(LMEM_FIXED, cbSize)) != NULL) {
		*pszBuffer = 0;
		cbSize--;
		for (; cbSize>0; pszLine += n, cbSize -= n, idStringNo++) {
			n = LoadString(hInst, idStringNo, (LPSTR)pszLine, cbSize);
			if (*(long *)pszLine == ASTASTAST) {
				*pszLine = 0;
				break;
			}
		}
	}
	else okMessage(STR_INSUFFICIENTHEAP);
	return pszBuffer;
#undef ASTASTAST
} // char *allocateAndLoad(int cbSize, UINT idStringNo)

void helpMessage(HWND hwnd, UINT idStringNo)
{
	char *pszBuffer;
	if ((pszBuffer = allocateAndLoad(512, idStringNo)) != NULL) {
		MessageBox(hwnd, pszBuffer, PROGNAME " Help", MB_OK);
		LocalFree((LOCALHANDLE)pszBuffer);
	}
} // void helpMessage(HWND hwnd, UINT idStringNo)

BOOL FAR PASCAL _export AboutDlgProc(HWND hwnd, WORD message,
	WORD wParam, LONG lParam)
{
	switch (message) {

	case WM_INITDIALOG:
		return TRUE;

	case WM_COMMAND:
		switch (wParam) {
			case IDOK:
			case IDCANCEL:
				EndDialog(hwnd, 0);
				return TRUE;
		}
		break;
	}
	return FALSE;
} // BOOL FAR PASCAL _export AboutDlgProc()

int HelpToClient(int x, int screen)
	{return (int)((((LONG)x * screen) + 512) / 1024);}

int ClientToHelp(int x, int screen)
	{return (int)((((LONG)x * 1024) + (screen >> 1)) / screen);}

BOOL CALLBACK _export WH_wSampleProc(HWND hwnd, WORD message,
	WORD wParam, LONG lParam)
{
#define BUFFER 2000
#define SMALLFONT 360
	static RECT		rectSL;
	static HWND		hwH, hwT, hwSL;
	static HBRUSH	hbH, hbT;
	static HFONT	hfH, hfT;
	static int		cx, cy, cw, ch;
	static BOOL		bMark = FALSE;
	char	*pszBuffer;
	int		i;
	BOOL	b;

	switch (message) {

	case WM_INITDIALOG:
		hwH = GetDlgItem(hwnd, ID_SAMPLE_HDR);
		hwT = GetDlgItem(hwnd, ID_SAMPLE_TXT);
		hwSL = GetDlgItem(hwnd, ID_SAMPLE_SL);
		GetClientRect(hwSL, &rectSL);
		if ((pszBuffer = allocateAndLoad(BUFFER, 1)) != NULL) {
			SetWindowText(hwT, pszBuffer);
			free(pszBuffer);
		}
		MoveWindow(hwnd, cx = HelpToClient(wsp->xCoord, cXscreen),
			cy = HelpToClient(wsp->yCoord, cYscreen),
			cw = HelpToClient(wsp->width, cXscreen),
			ch = HelpToClient(wsp->height, cYscreen), TRUE);
		hbH = wsp->flag & WS_NSCOLOR ?
			CreateSolidBrush(wsp->nonscrollColor) : 0;
		hbT = wsp->flag & WS_SCOLOR ?
			CreateSolidBrush(wsp->scrollColor) : 0;
		logfont.lfHeight = !bSampleMode || cw > SMALLFONT ? 24 : 16;
		strcpy(logfont.lfFaceName, "Times New Roman");
		hfH = CreateFontIndirect(&logfont);
		logfont.lfHeight = 15;
		strcpy(logfont.lfFaceName, "Courier New");
		hfT = CreateFontIndirect(&logfont);
		SendMessage(GetDlgItem(hwnd, IDC_SAMPLE_SL), BM_SETCHECK, bMark, 0L);
		EnableWindow(GetDlgItem(hwnd, IDOK), bSampleMode);
		SetDlgItemText(hwnd, ID_SAMPLE_HDR,
			!bSampleMode ? "  INT 09 - IRQ1 - KEYBOARD DATA READY" :
			cw > SMALLFONT ? "  Position sample window and click on OK." :
			"  Position sample window\n   and click on OK.");
	setSLwindow:
		ShowWindow(hwSL, bMark ? SW_SHOWNORMAL : SW_HIDE);
		return TRUE;

	case WM_MOVE:
		cXorig = LOWORD(lParam) - cXoffs;
		cYorig = HIWORD(lParam) - cYoffs;
		MoveWindow(hwSL, 20, 480 - cYoffs - cScanmark,
			rectSL.right, rectSL.bottom, TRUE);
		return TRUE;

	case WM_SIZE:
		cWidth = LOWORD(lParam) + cXadj;
		cHeight = HIWORD(lParam) + cYadj;
		return TRUE;

	case WM_COMMAND:
		switch (wParam) {
			case IDC_SAMPLE_SL:
				bMark = (BOOL)SendMessage(GetDlgItem(hwnd, IDC_SAMPLE_SL),
					BM_GETCHECK, 0, 0L);
				goto setSLwindow;

			case IDOK:
				if (cx != cXorig)
				// copy only if changed to avoid rounding errors from
				// transformation forth and back
					wsp->xCoord = ClientToHelp(cXorig, cXscreen);
				if (cy != cYorig)
					wsp->yCoord = ClientToHelp(cYorig, cYscreen);
				if (cw != cWidth)
					wsp->width = ClientToHelp(cWidth, cXscreen);
				if (ch != cHeight)
					wsp->height = ClientToHelp(cHeight, cYscreen);
				iNeedSave |= SAVEWINDOW;
				ewp->changed |= PSOK;
				b = FALSE;
				if ((int)(wsp->xCoord) < 0) {
					wsp->xCoord = 0;
					b = TRUE;
				}
				if ((int)(wsp->yCoord) < 0) {
					wsp->yCoord = 0;
					b = TRUE;
				}
				if ((i = wsp->xCoord + wsp->width) > 1023) {
					if (i>1024) {
						wsp->width = 1023 - wsp->xCoord;
						b = TRUE;
					}
					else wsp->width--;
				}
				if ((i = wsp->yCoord + wsp->height) > 1023) {
					if (i>1024) {
						wsp->height = 1023 - wsp->yCoord;
						b = TRUE;
					}
					else wsp->height--;
				}
				if (b)
					okMessage(STR_CLIPPED);
					// fall through
			case IDCANCEL:
				if (hbH) DeleteObject(hbH);
				if (hbT) DeleteObject(hbT);
				DeleteObject(hfT);
				DeleteObject(hfH);
				EndDialog(hwnd, 0);
				return TRUE;

			case ID_SAMPLE_HELP:
				helpMessage(hwnd, bSampleMode ? 400 : 416);
				return TRUE;
			}

		case WM_CTLCOLOR:
			if ((HWND)LOWORD(lParam) == hwH) {
				SetBkMode((HDC)wParam, TRANSPARENT);
				SelectObject((HDC)wParam, hfH);
				return hbH;
			}
			if ((HWND)LOWORD(lParam) == hwT) {
				SetBkMode((HDC)wParam, TRANSPARENT);
				SelectObject((HDC)wParam, hfT);
				return hbT;
			}
			break;

	}
	return FALSE;
#undef BUFFER
#undef SMALLFONT
} // BOOL CALLBACK _export WH_wSampleProc()

BOOL atou(LPSTR *plpPar, unsigned *puVal)
{
	LPSTR lp;
	char c;
	while (**plpPar == ' ')
		(*plpPar)++;
	*puVal = 0;
	for (lp = *plpPar; (c = **plpPar) >= '0' && c <= '9'; (*plpPar)++)
		*puVal = *puVal * 10 + (c-'0');
	return *plpPar != lp;
} // BOOL atoint()

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
				LPSTR lpszCmdLine, int nCmdShow)
{
	MSG			msg;
	WNDCLASS	wndclass;
	HDC			hdc;

	hInst = hInstance;

	if (!hPrevInstance) {
		wndclass.style			= 0;
		wndclass.lpfnWndProc	= WndProc;
		wndclass.cbClsExtra		= 0;
		wndclass.cbWndExtra		= DLGWINDOWEXTRA;
		wndclass.hInstance		= hInstance;
		wndclass.hIcon			= LoadIcon(hInstance, szAppName);
		wndclass.hCursor		= LoadCursor(NULL, IDC_ARROW);
		wndclass.hbrBackground	= COLOR_WINDOW + 1;
		wndclass.lpszMenuName	= szAppName;
		wndclass.lpszClassName	= szAppName;

		RegisterClass(&wndclass);

		wndclass.lpfnWndProc	= EditProc;
		wndclass.cbWndExtra		= 0;
		wndclass.lpszMenuName	= "EditMacro";
		wndclass.lpszClassName	= szEditFrame;

		RegisterClass(&wndclass);

	}

	ofn.hwndOwner = cc.hwndOwner = hwndMain =
		CreateDialog(hInstance, szAppName, 0 , NULL);
	hwndHeaderSample = GetDlgItem(hwndMain, IDT_HEADER_SAMPLE);
	hwndTextSample = GetDlgItem(hwndMain, IDT_TEXT_SAMPLE);
	idColorHelp = RegisterWindowMessage(HELPMSGSTRING);

	lpfnAboutDlgProc = MakeProcInstance((FARPROC)AboutDlgProc, hInstance);
	lpfnWH_wSampleProc = MakeProcInstance((FARPROC)WH_wSampleProc, hInstance);

	atou(&lpszCmdLine, (unsigned *)&cXorig);
	atou(&lpszCmdLine, (unsigned *)&cYorig);
	atou(&lpszCmdLine, (unsigned *)&cWidth);
	if ((bEditEnabled = atou(&lpszCmdLine, (unsigned *)&cHeight)) == FALSE)
		{cXorig = 100; cYorig = 20; cWidth = 530; cHeight = 280;}
	hwndEditFrame = CreateWindow(szEditFrame,
		PROGNAME " Macros Editor",
		WS_OVERLAPPED | WS_CAPTION | WS_THICKFRAME |
		WS_MINIMIZEBOX | WS_MAXIMIZEBOX,
		cXorig, cYorig, cWidth, cHeight,
		NULL, NULL, hInst, NULL);

	if (bEditEnabled)
		ShowWindow(hwndEditFrame, SW_SHOWNORMAL);

	ShowWindow(hwndMain, nCmdShow);
	UpdateWindow(hwndMain);

	hdc = GetDC(hwndMain);
	GetTextMetrics(hdc, &tm);
	ReleaseDC(hdc, hdc);
	cScanmark = (tm.tmHeight + tm.tmExternalLeading) * 3 - tm.tmDescent;
	cXscreen = GetSystemMetrics(SM_CXSCREEN);
	cYscreen = GetSystemMetrics(SM_CYSCREEN);
	cYcaption = GetSystemMetrics(SM_CYCAPTION);
	cXdlgframe = GetSystemMetrics(SM_CXDLGFRAME);
	cYdlgframe = GetSystemMetrics(SM_CYDLGFRAME);
	cYmenu = GetSystemMetrics(SM_CYMENU);
	cXoffs = cXdlgframe;
	cYoffs = cYcaption + cYdlgframe + cYmenu;
	cXadj = cXoffs + cXdlgframe;
	cYadj = cYoffs + cYdlgframe;

	while (GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.wParam;
} // int PASCAL WinMain()

void DoCaption(void)
{
	char szCaption[64 + _MAX_FNAME + _MAX_EXT];

	wsprintf(szCaption, "%s%s%s", (LPSTR) szAppName,
			(LPSTR)(szTitleName[0] ? ": " : ""),
			(LPSTR)szTitleName);

	SetWindowText(hwndMain, szCaption);
} // void DoCaption(void)

void AskAboutCopy(void)
/* copies editor contents to clipboard on request */
{
	if (iNeedSave && GetWindowTextLength(hwndEdit) &&
			MessageBox(hwndMain, "Copy Editor contents to clipboard?",
			szAppName, MB_YESNO | MB_ICONQUESTION) == IDYES) {
		SendMessage(hwndEdit, EM_SETSEL, 0, MAKELONG(0, 0xFFFE));
		SendMessage(hwndEdit, WM_COPY, 0, 0L);
	} // if (bNeedSave ...)
} // void AskAboutCopy(void)

BOOL Cancel(void)
{
	char  szBuffer[64 + _MAX_FNAME + _MAX_EXT];
	short nReturn;

	if (!iNeedSave)
		return FALSE;
	if (!bFileLoaded) {
		AskAboutCopy();
		return FALSE;
	} // if (!bFileLoaded)
	iSave = 0;
	wsprintf(szBuffer, "Save WINDOWS changes in %s?", (LPSTR)szTitleName);
	if ((iNeedSave & SAVEWINDOW) && (nReturn = MessageBox(hwndMain, szBuffer,
			szAppName, MB_YESNOCANCEL | MB_ICONQUESTION)) == IDYES)
		iSave |= SAVEWINDOW;
	if (nReturn == IDCANCEL)
		return FALSE;
	wsprintf(szBuffer, "Save MACROS changes in %s?", (LPSTR)szTitleName);
	if ((iNeedSave & SAVEMACRO) && (nReturn = MessageBox(hwndMain, szBuffer,
			szAppName, MB_YESNOCANCEL | MB_ICONQUESTION)) == IDYES)
		iSave |= SAVEMACRO;
	if (iSave)
		if (SendMessage(hwndMain, WM_COMMAND, IDM_SAVE, 0L))
			nReturn = IDCANCEL;
	return nReturn == IDCANCEL;
} // BOOL Cancel(void)

void initFlags(void)
{
	int i;
	EXT_WS *ewp;
	for (i=0; i<nCount; i++) {
		ewp = (EXT_WS *)LocalLock(ewpWindows[i]);
		ewp->changed = 0;
		ewp->flag = ewp->w.flag;
		ewp->pos = *(LONG*)&(ewp->w.xCoord);
		ewp->size = *(LONG*)&(ewp->w.width);
		if ((ewp->w.xCoord | ewp->w.yCoord | ewp->w.width |
				ewp->w.height | (ewp->w.xCoord + ewp->w.width) |
				(ewp->w.yCoord + ewp->w.height)) > 1023) {
			ewp->w.xCoord = 0;
			ewp->w.yCoord = 0;
			ewp->w.width = 980;
			ewp->w.height = 1023;
		}
		ewp->rgbNSC = ewp->w.nonscrollColor;
		ewp->rgbSC = ewp->w.scrollColor;
		LocalUnlock(ewpWindows[i]);
	}
	iNeedSave = iSave = 0;
} // void initFlags(void)

void enableAll(BOOL b)
{
	int i;
	for (i=0; i<elements(toEnable); i++)
		EnableWindow(GetDlgItem(hwndMain, toEnable[i]), b);
} // void enableAll(BOOL b)

void showWindow(UINT w, BOOL b)
{
	HWND hwnd1;
	ShowWindow(hwnd1 = GetDlgItem(hwndMain, w), b ? SW_SHOWNORMAL : SW_HIDE);
	InvalidateRect(hwnd1, NULL, TRUE);
} // void showWindow(UINT w, BOOL b)

void unlockCurrent(void)
{
	if (iCurrentWindow != -1) {
		LocalUnlock(ewpWindows[iCurrentWindow]);
		iCurrentWindow = -1;
	} // if (iCurrentWindow != -1)
} // void unlockCurrent(void)

void freeWindows(void)
{
	int i;
	unlockCurrent();
	for (i=nCount-1; i>=0; i--)
		LocalFree(ewpWindows[i]);
	return;
} // void freeWindows(void)

void clearWindows(void)
{
	int i;
	freeWindows();
	enableAll(FALSE);
	for (i=0; i<elements(textWindows); i++)
		SetDlgItemText(hwndMain, textWindows[i], (LPSTR)"");
	for (i=0; i<elements(cbWindows); i++)
		SendDlgItemMessage(hwndMain, cbWindows[i], BM_SETCHECK, FALSE, 0L);
	for (i=0; i<elements(toHide); i++)
		showWindow(toHide[i], FALSE);
	if (!nCount)
		nCount = 1;
	while (nCount)
		SendDlgItemMessage(hwndMain, IDL_WINDOW, LB_DELETESTRING, --nCount, 0L);
	SetWindowText(hwndMain, (LPSTR)szAppName);
	SetWindowText(hwndEdit, (LPSTR)"");
	bFileLoaded = FALSE;
	return;
} // void clearWindows(void)

void errorWMessage(int nMessage)
{
	close(hFile);
	okMessage(nMessage);
} // void errorWMessage(int nMessage)

void errorMessage(int nMessage)
{
	freeWindows();
	errorWMessage(nMessage);
} // void errorMessage(int nMessage)

void testCheckbox(WORD mask, UINT change)
{
	BOOL b;
	b = (wsp->flag & mask)!=0;
	SendDlgItemMessage(hwndMain, change+1, BM_SETCHECK, b, 0L);
	EnableWindow(GetDlgItem(hwndMain, change+2), b);
	if (change == IDC_POS_SIZE_CHANGE) {
		showWindow(change, TRUE);
		SetDlgItemText(hwndMain, change, b ? "Change" : "Test");
		bSampleMode = b;
	}
	else {
		showWindow(change, b);
		showWindow(change+3, b);
	}
} // void testCheckbox(WORD mask, UINT change)

void writePosSize(void)
{
	wsprintf(buffer, "%u, %u, %u, %u", wsp->xCoord, wsp->yCoord,
		wsp->width, wsp->height);
	SetDlgItemText(hwndMain, IDT_POS_SIZE, (LPSTR)buffer);
} // void writePosSize(void)

void writeRGB(WORD mask, COLORREF rgb, UINT change)
{
	testCheckbox(mask, change);
	wsprintf(buffer, "%u, %u, %u", GetRValue(rgb), GetGValue(rgb),
		GetBValue(rgb));
	SetDlgItemText(hwndMain, change+2, (LPSTR)buffer);
} // void writeRGB(WORD mask, COLORREF rgb, UINT change)


void updateDlg(void)
{
	unlockCurrent();
	iCurrentWindow = (int)
		SendDlgItemMessage(hwndMain, IDL_WINDOW, LB_GETCURSEL, 0, 0L);
	ewp = (EXT_WS *)LocalLock(ewpWindows[iCurrentWindow]);
	wsp = &(ewp->w);
	enableAll(TRUE);
	SetDlgItemText(hwndMain, IDT_NAME, (LPSTR)wsp->name);
	SetDlgItemText(hwndMain, IDT_KIND, (LPSTR)wsp->kind);
	SetDlgItemText(hwndMain, IDT_TITLE, (LPSTR)(wsp->flag & WS_TITLE ?
		wsp->title : "(no title in [Windows] section)"));
	testCheckbox(WS_POSSIZE, IDC_POS_SIZE_CHANGE);
	EnableWindow(GetDlgItem(hwndMain, IDC_POS_SIZE), !ewp->secondary);
	writePosSize();
	writeRGB(WS_NSCOLOR, wsp->nonscrollColor, IDC_HEADER_CHANGE);
	writeRGB(WS_SCOLOR, wsp->scrollColor, IDC_TEXT_CHANGE);
	return;
} // void updateDlg(void)

void whweReadWindows(void)
{
	int i, comp, err, iInMacros;
	BOOL bInMacros;
	SYSTEM sys;
	LOCALLINK ll;
	struct INFO {
		DWORD length;
		DWORD datalength;
		unsigned char flag;
		WORD info;
	} info;
	DWORD dwFrom, dwTo;
	struct date compDate;
	struct time compTime;
	char *pszM, *pszMv;
	EXT_WS *ewp;
#define BUFFER 514
		// macro strings can be up to 512 characters long
	union buffer {
		char str[BUFFER];
		WINDOWSTRUCT ws;
		DWORD ul;
	} *pBuffer;
	clearWindows();
	if ((pBuffer = (union buffer NEAR *)LocalAlloc(LPTR, BUFFER)) == NULL) {
		errorMessage(STR_INSUFFICIENTHEAP);
		return;
	}
	if ((err=openHelpFile(&hFile, szFileName, &root, WRITEACCESS))!=0) {
		if (err == HBERR_OPEN)
			okMessage(STR_CANNOTOPEN);
		else
			errorMessage(STR_UNKNOWNFORMAT);
		
		goto exit;
	}
	if ((err = findIndexTableEntry(hFile, "|SYSTEM", &dwSysStart)) != 0) {
		errorMessage(err == HBERR_NO_ENTRY ? STR_CANTFINDSYSTEM :
			err == HBERR_CANT_ALLOC ? STR_INSUFFICIENTHEAP : STR_INDEXERROR);
		goto exit;
	}
	if (readBlock(hFile, dwSysStart, &info, sizeof(info))) {
		errorMessage(STR_SYSTEMERROR);
		goto exit;
	} /* if (readBlock()) */
	if (info.flag != 0 || info.info != 0x036C) {
		errorMessage(STR_UNKNOWNSYSTEM);
		goto exit;
	} /* if (info.flag != 0 || info.type != 0x036C) */
	if (readBlock(hFile, dwSysStart + 9, &sys, sizeof(sys))) {
		errorMessage(STR_BADSYSTEM);
		goto exit;
	} /* if (readBlock()) */
	if (sys.type!=0x036C || (sys.c1!=0x0F &&  sys.c1!=0x15) ||
			sys.c2!=0 || sys.c3!=1 || sys.c4!=0) {
		errorMessage(STR_SYSTEMSTRUCTURE);
		goto exit;
	} /* if (sys.type!=0x036C ...) */
	unixtodos(sys.time-3*3600, &compDate, &compTime);
		/* HC times seem to be 3 hours after actual compilation time */
	wsprintf(buffer, "%u-%02u-%02u - %u:%02u",
		compDate.da_year, compDate.da_mon, compDate.da_day,
		compTime.ti_hour, compTime.ti_min);
	if (sys.c1==0x0F) {
		errorMessage(STR_HC30);
		goto exit;
	} /* if (sys.c1==0x0F) */
	if ((pszM = pszMv = (char NEAR *)LocalAlloc(LPTR, (int)info.length)) == NULL) {
		errorMessage(STR_INSUFFICIENTHEAP);
		goto exit;
	}
	dwFrom = dwSysStart + 9 + sizeof(sys);
	dwSysEnd = dwTo = dwSysStart + info.datalength + 9;
	dwSysLength = info.length;
	dwSysMacros = dwSysWindows = 0;
	iInMacros = 0;
	bInMacros = FALSE;
	while (dwFrom < dwTo) {
		if (readBlock(hFile, dwFrom, &ll, sizeof(ll)) ||
				readBlock(hFile, dwFrom+sizeof(ll), pBuffer,
				(ll.length<BUFFER ? ll.length : BUFFER - 1))) {
			errorMessage(STR_SYSTEMTABLE);
			goto exit;
		} /* if (err) */
		if (ll.type != MACRO_TYPE)
			bInMacros = FALSE;
		switch (ll.type) {
			case MACRO_TYPE:
				if (dwSysMacros == 0)
					dwSysMacros = dwFrom;
				if (!bInMacros) {
					iInMacros++; // happens only once if macros are contiguous
					bInMacros = TRUE;
				}
				strcpy(pszMv, pBuffer->str);
				pszMv += strlen(pszMv);
				*pszMv++ = '\r';
				*pszMv++ = '\n';
				break;
			case WINDOW_TYPE:
				if (dwSysWindows == 0)
					dwSysWindows = dwFrom;
				if (ll.length != sizeof(WINDOWSTRUCT)) {
					errorMessage(STR_WINDOWSTRUCT);
					goto exit;
				}
				if (nCount >= MAX_WINDOWS) {
					errorMessage(STR_TOOMANY);
					goto exit;
				}
				if ((ewp = (EXT_WS *)LocalLock(ewpWindows[nCount] =
						LocalAlloc(LMEM_MOVEABLE, sizeof(EXT_WS))))==NULL) {
					errorMessage(STR_INSUFFICIENTHEAP);
					goto exit;
				}
				ewp->offset = (WORD)(dwFrom - dwSysWindows) + sizeof(ll);
				ewp->length = ll.length;
				ewp->secondary=!strcmp(pBuffer->ws.kind,"secondary");
				movmem(pBuffer, &(ewp->w), ll.length);
				LocalUnlock(ewpWindows[nCount]);
	
				nCount++;
				if (((i=pBuffer->ws.flag & WS_POSSIZE) != WS_POSSIZE) && i) {
					errorMessage(STR_UNKNOWNFLAG);
					goto exit;
				}
				break;
			case TITLE_TYPE:
			case COPYRIGHT_TYPE:
			case CONTENTS_TYPE:
			case ICON_TYPE:
			case CITATION_TYPE:
				break;
			default:
				errorMessage(STR_SYSTEMENTRY);
				goto exit;
		} // switch(ll.type)
		dwFrom += (ll.length+sizeof(ll));
	} /* while (dwFrom < dwTo) */
	bStrangeMacro = iInMacros > 1;
	if (bStrangeMacro) // macros are not contiguous
		okMessage(STR_STRANGEMACRO);
	SetWindowText(hwndEdit, pszM);
	if (!dwSysWindows)
		dwSysWindows = dwSysEnd;
	if (!dwSysMacros)
		dwSysMacros = dwSysWindows;
	initFlags();
	SetDlgItemText(hwndMain, IDT_FILENAME, (LPSTR)szTitleName);
	SetDlgItemText(hwndMain, IDT_COMPILED, (LPSTR)buffer);
	if (nCount) {
		for (i=0; i<nCount; i++) {
			ewp = (EXT_WS *)LocalLock(ewpWindows[i]);
			SendDlgItemMessage(hwndMain, IDL_WINDOW, LB_INSERTSTRING,
				i, (LPARAM)(LPCSTR)&(ewp->w.name));
			LocalUnlock(ewpWindows[i]);
		} // for (i=0; ...
		SendDlgItemMessage(hwndMain, IDL_WINDOW, LB_SETCURSEL, 0, 0L);
		updateDlg();
	}
	else {
		okMessage(STR_NOWINDOWS);
		SendDlgItemMessage(hwndMain, IDL_WINDOW, LB_INSERTSTRING,
			0, (LPARAM)(LPCSTR)"(none defined)");
		if (bEditEnabled)
			SetFocus(hwndEdit);
	} // else
	close(hFile);
	DoCaption();
	bFileLoaded = TRUE;
exit:
	LocalFree((LOCALHANDLE)pBuffer);
	LocalFree((LOCALHANDLE)pszM);
	return;
#undef BUFFER
} // void whweReadWindows(void)

BOOL whweSaveWindows(void)
{
	int i, nLines;
	BOOL bAtEnd, bRelease, bReturn = TRUE;
	WORD wILength, wMLength, wWLength, wLength;
	DWORD dwNewOffset, dwSavedLength, dwUnused;
	char *pszBuffer = NULL, *pbBuffer, *psz1, *psz2;
	EXT_WS *ewp;
	LOCALLINK ll;
	if ((hFile=open(szFileName, O_RDWR|O_BINARY)) < 0) {
		okMessage(STR_CANNOTREOPEN);
		return FALSE;
	}
	if (lseek(hFile, 0L, SEEK_END) != root.length) {
		errorWMessage(STR_WRONGLENGTH);
		return FALSE;
	}

	if (iSave & SAVEWINDOW)
		for (i=0; i<nCount; i++) {
			ewp = (EXT_WS *)LocalLock(ewpWindows[i]);
			if (ewp->changed) {
				if (!(ewp->changed & PSOK)) { // not changed, restore orig. pos.
					*(LONG*)&(ewp->w.xCoord) = ewp->pos;
					*(LONG*)&(ewp->w.width) = ewp->size;
				} // if (!...)
				if (writeBlock(hFile, dwSysWindows + ewp->offset, &(ewp->w),
						ewp->length)) {
					errorWMessage(STR_ERRORUPDATE);
					return FALSE;
				} // if (writeBlock(...)
			}
			LocalUnlock(ewpWindows[i]);
		} // for (i=0;...)

	if (iSave & SAVEMACRO) {
		if (bStrangeMacro) {
			okMessage(STR_STRANGEMACRO);
			AskAboutCopy();
			goto endOfMacros;
		} // if (bStrangeMacro)
		wILength = (WORD)(dwSysMacros - dwSysStart);
		wWLength = (WORD)(dwSysEnd - dwSysWindows);
		pbBuffer = (char NEAR *)LocalAlloc(LPTR, wILength > wWLength ?
			wILength : wWLength);
		if (pbBuffer == NULL) {
			errorWMessage(STR_INSUFFICIENTHEAP);
			return FALSE;
		} // if (pbBuffer == NULL)
		wMLength = GetWindowTextLength(hwndEdit);
		nLines = 0;
		if (wMLength) {
			psz1 = pszBuffer = (char NEAR *)LocalAlloc(LPTR, wMLength+3);
			if (pszBuffer == NULL) {
				okMessage(STR_INSUFFICIENTHEAP);
				AskAboutCopy();
				bReturn = FALSE;
				goto releasePbBuffer;
			} // if (wMLength && ...)
			GetWindowText(hwndEdit, pszBuffer, wMLength+1);
			if (*(psz1 + wMLength - 1) != '\n') {
				*(psz1 + wMLength++) = '\r'; // terminate last line
				*(psz1 + wMLength++) = '\n';
				*(psz1 + wMLength) = 0;
			}
			psz2 = psz1 + wMLength;
			for (; psz1 < psz2; psz1++)
				if (*psz1 == '\r')
					nLines++; // count lines
			wMLength += ((5-2) * nLines);
			// line overhead is 2 in editor and 5 in file
			// (wMLength is now the length of the final Macros structure)
		} // if (wMLength)
		wLength = wILength + wMLength + wWLength;
		bAtEnd = dwSysStart + dwSysLength == root.length;
		dwUnused = 0;
		bRelease = FALSE;
		dwSavedLength = wLength;
		if (bAtEnd || wLength > dwSysLength)
			 findUnusedBlock(hFile, &dwSavedLength, &dwUnused);
		if (bAtEnd) {
			if (dwUnused) { // use the unused area
				root.length = dwSysStart;
				goto useUnused;
			} // if (dwUnused)
			else {
				root.length -= (dwSysLength - wLength);
				dwSavedLength = wLength;
				goto useSamePos;
			} // else
		} // if (bAtEnd)
		if (wLength > dwSysLength) {
			// insufficient local space, find new file location for block
			bRelease = TRUE; // release current block after transfer
			if (dwUnused) // use the unused area
		useUnused:
				dwNewOffset = dwUnused;
			else { // append to end of file
				dwNewOffset = root.length;
				root.length += wLength;
			} // else
		} // if (wLength > dwSysLength)
		else {
			dwSavedLength = dwSysLength;
	useSamePos:
			dwNewOffset = dwSysStart;
		} // else
		writeBlock(hFile, 12, &(root.length), 4);
			// update file's length
		writeBlock(hFile, getLatestInfoDataOffset(), &dwNewOffset, 4);
			// insert (new?) offset to |SYSTEM block

		// transfer initial part
		readBlock(hFile, dwSysStart, pbBuffer, wILength);
		((BLOCKINFO *)pbBuffer)->datalength = wLength - 9;
			// wLength includes the size of the link (9 bytes)
		((BLOCKINFO *)pbBuffer)->length = dwSavedLength;
		writeBlock(hFile, dwNewOffset, pbBuffer, wILength);

		if (wWLength) { // read Windows info
			readBlock(hFile, dwSysWindows, pbBuffer, wWLength);
			lseek(hFile, dwNewOffset + wILength, SEEK_SET);
		} // if (wWLength)
		if (wMLength) { // write Macros info
			psz1 = pszBuffer;
			ll.type = MACRO_TYPE;
			for (i = 0; i < nLines; i++) {
				for (psz2 = psz1; *psz2 != '\r'; psz2++)
					;
				*psz2 = 0;
				ll.length = psz2 - psz1 + 1;
				write(hFile, &ll, sizeof(ll));
				write(hFile, psz1, ll.length);
				psz1 = psz2 + 2;
			} // for (i = 0; ...)
		} // if (wMLength)
		if (wWLength)
			write(hFile, pbBuffer, wWLength);
		dwSysMacros = dwNewOffset + wILength;
		dwSysWindows = dwSysMacros + wMLength;
		dwSysEnd = dwSysWindows + wWLength;
		dwSysLength = dwSavedLength;
		if (tell(hFile) != dwSysEnd) {
			okMessage(STR_FILECORRUPTED);
			AskAboutCopy();
		} // if (tell(hFile) != dwSysEnd)
		if (bRelease) {
			linkUnusedBlock(hFile, dwSysStart);
			readHelpRoot(hFile, &root); // update although not used
		} // if (bRelease)
		dwSysStart = dwNewOffset;
		chsize(hFile, root.length); // set file length
	releaseHeap:
		LocalFree((LOCALHANDLE)pszBuffer);
	releasePbBuffer:
		LocalFree((LOCALHANDLE)pbBuffer);
endOfMacros:
	} // if (iSave & SAVEWINDOW)
	close(hFile);
	initFlags();
	return bReturn;
} // BOOL whweSaveWindows(void)

long CALLBACK _export EditProc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
{
	LONG lSelect;
	WORD wEnable;
	switch (message) {
		case WM_CREATE:
			hwndEdit = CreateWindow("edit", "",
				WS_CHILD | WS_VISIBLE | WS_HSCROLL | WS_VSCROLL |
				ES_LEFT | ES_MULTILINE | ES_AUTOHSCROLL | ES_AUTOVSCROLL,
				0, 0, 0, 0,
				hwnd, EDITID, hInst, NULL);
			return 0;
 
		case WM_SIZE:
			MoveWindow(hwndEdit, 0, 0, LOWORD(lParam), HIWORD(lParam), TRUE);

		case WM_SETFOCUS:
			SetFocus(hwndEdit);
			return 0;

		case WM_INITMENUPOPUP:
			if (lParam == 1) { // Edit menu

				// Enable Undo if edit control can do it
				EnableMenuItem(wParam, IDM_UNDO,
					SendMessage(hwndEdit, EM_CANUNDO, 0, 0L) ?
					MF_ENABLED : MF_GRAYED);

				// Enable Paste if text is in the clipboard
				EnableMenuItem(wParam, IDM_PASTE,
					 IsClipboardFormatAvailable(CF_TEXT) ?
					 MF_ENABLED : MF_GRAYED);

				// Enable Cut, Copy, and Del if text is selected

				lSelect = SendMessage(hwndEdit, EM_GETSEL, 0, 0L);
				wEnable = HIWORD(lSelect) != LOWORD(lSelect) ?
					MF_ENABLED : MF_GRAYED;

				EnableMenuItem (wParam, IDM_CUT,  wEnable);
				EnableMenuItem (wParam, IDM_COPY, wEnable);
				EnableMenuItem (wParam, IDM_DEL,  wEnable);
			}
			break;

		case WM_COMMAND:

			// Messages from edit control
			if (LOWORD(lParam) && (wParam == EDITID)) {
				if ((HIWORD(lParam) == EN_UPDATE)) {
					iNeedSave |= SAVEMACRO;
					return 0;
				}
				if (HIWORD(lParam) == EN_ERRSPACE || HIWORD(lParam) == EN_MAXTEXT) {
					okMessage(STR_INSUFFICIENTHEAP);
					return 0;
				}
				break;
			}

			// Messages from File menu
			switch (wParam) {
				case IDM_MAIN:
					ShowWindow(hwndMain, SW_SHOWNORMAL);
					SetFocus(hwndMain);
					return 0;

				case IDM_UNDO:
					SendMessage(hwndEdit, WM_UNDO, 0, 0L);
					return 0 ;

				case IDM_CUT:
					SendMessage(hwndEdit, WM_CUT, 0, 0L);
					return 0 ;

				case IDM_COPY:
					SendMessage(hwndEdit, WM_COPY, 0, 0L);
					return 0 ;

				case IDM_PASTE:
					SendMessage(hwndEdit, WM_PASTE, 0, 0L);
					return 0 ;

				case IDM_DEL:
					SendMessage(hwndEdit, WM_CLEAR, 0, 0L);
					return 0 ;

				case IDM_HELP:
	showHelp:
					helpMessage(hwnd, 448);
					return 0;
			}
			break;

		case WM_KEYDOWN:
			if (wParam == VK_F1)
				goto showHelp;
			break;

		case WM_CLOSE:
			return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
} // long CALLBACK _export EditProc()

long CALLBACK _export WndProc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
{
	static HBRUSH hbrush;
	BOOL b;

	if (message == idColorHelp)
		helpMessage(hwnd, 432);
	else switch (message) {
		case WM_INITMENUPOPUP:
			if (HIWORD(lParam)) { /* disable Size and Maximize */
				EnableMenuItem(wParam, 2, MF_BYPOSITION | MF_GRAYED);
				EnableMenuItem(wParam, 4, MF_BYPOSITION | MF_GRAYED);
			}
			return 0;

		case WM_COMMAND:
			switch (wParam) {
				case IDM_OPEN:
					if (Cancel())
						return 0;
					ofn.Flags = OFN_FILEMUSTEXIST;
					if (GetOpenFileName(&ofn))
						whweReadWindows();
					return 0;

				case IDM_SAVE:
					if (!iNeedSave) {
						okMessage(STR_NOTHINGTOSAVE);
						return 0;
					}
					if (!bFileLoaded) {
						AskAboutCopy();
						return 0;
					}
					if (!iSave)
						iSave = iNeedSave;
					return !whweSaveWindows();

				case IDM_EXIT:
					SendMessage (hwnd, WM_CLOSE, 0, 0L);
					return 0 ;

				case IDL_WINDOW:
					if (HIWORD(lParam) == CBN_SELCHANGE)
						updateDlg();
					return 0 ;

				case IDC_POS_SIZE:
					wsp->flag ^= WS_POSSIZE;
					testCheckbox(WS_POSSIZE, IDC_POS_SIZE_CHANGE);
					iNeedSave |= SAVEWINDOW;
					ewp->changed |= PSDEF;
					return 0 ;

				case IDC_POS_SIZE_CHANGE:
					DialogBox(hInst, "WH_wSample", hwnd, lpfnWH_wSampleProc);
					writePosSize();
					return 0 ;

				case IDC_HEADER_COLOR:
					wsp->flag ^= WS_NSCOLOR;
					ewp->changed |= HDEF;
			writeHEADER:
					writeRGB(WS_NSCOLOR, wsp->nonscrollColor,
						IDC_HEADER_CHANGE);
					iNeedSave |= SAVEWINDOW;
					return 0 ;

				case IDC_TEXT_COLOR:
					wsp->flag ^= WS_SCOLOR;
					ewp->changed |= TDEF;
			writeTEXT:
					writeRGB(WS_SCOLOR, wsp->scrollColor,
						IDC_TEXT_CHANGE);
					iNeedSave |= SAVEWINDOW;
					return 0 ;

				case IDC_HEADER_CHANGE:
					cc.rgbResult = wsp->nonscrollColor;
					rgbUser[15] = ewp->rgbNSC;
					if (ChooseColor(&cc)) {
						ewp->changed |= HOK;
						wsp->nonscrollColor = cc.rgbResult;
						goto writeHEADER;
					}
					return 0;

				case IDC_TEXT_CHANGE:
					cc.rgbResult = wsp->scrollColor;
					rgbUser[15] = ewp->rgbSC;
					if (ChooseColor(&cc)) {
						ewp->changed |= TOK;
						wsp->scrollColor = cc.rgbResult;
						goto writeTEXT;
					}
					return 0;

				case IDM_EDIT:
					if (!bEditEnabled && (bEditEnabled = (MessageBox(hwnd,
							"WARNING!\n\n"
							"The Macros Editor lets you edit the macros defined"
							" in the Help Project File's [CONFIG] section.\n\n"
							"Macro editing is a delicate matter. Do NOT use"
							" the Macros Editor unless you know what you are"
							" doing.\n\n"
							"Enter the Macros Editor?", szAppName,
							MB_YESNO | MB_ICONEXCLAMATION) == IDYES)) != 0)
						ShowWindow(hwndEditFrame, SW_SHOWNORMAL);
					if (bEditEnabled) {
						if (IsIconic(hwndEditFrame))
							ShowWindow(hwndEditFrame, SW_SHOWNORMAL);
						UpdateWindow(hwndEditFrame);
						SetFocus(hwndEditFrame);
					}
					return 0;

				case IDM_ABOUT:
	showAbout:
					DialogBox(hInst, "AboutBox", hwnd, lpfnAboutDlgProc);
					return 0;
			}
			break;

		case WM_KEYDOWN:
			if (wParam == VK_F1)
				goto showAbout;
			break;

		case WM_CTLCOLOR:
			if ((b = ((HWND)LOWORD(lParam) == hwndHeaderSample)) != 0 ||
					(HWND)LOWORD(lParam) == hwndTextSample) {
				if (hbrush)
					DeleteObject(hbrush);
				return hbrush = CreateSolidBrush(b ?
					wsp->nonscrollColor : wsp->scrollColor);
			}
			break;

		case WM_CLOSE:
			if (!Cancel())
				DestroyWindow(hwnd);
			return 0;

		case WM_QUERYENDSESSION:
			return Cancel() ? 0L : 1L;

		case WM_DESTROY:
			DestroyWindow(hwndEdit);
			if (hbrush)
				DeleteObject(hbrush);
			PostQuitMessage(0);
			return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
} // long CALLBACK _export WndProc()
